/*
*  Copyright (C) 1998-2022 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
  Test class
  TestCollection class
  TestRoot global variable
  initTree function
*/

// This code assumes there's a "myDiagram" div for test diagrams,
// a "myTestTree" div for holding the tree of test cases,
// and a "myStatus" div for showing status messages.
var isMac = (this.navigator !== undefined && this.navigator.platform !== undefined && this.navigator.platform.toUpperCase().indexOf('MAC') >= 0);

(function(factory) {
  if (typeof module === "object" && typeof module.exports === "object") {
    var v = factory(require, exports);
    if (v !== undefined) module.exports = v;
  } else if (typeof define === "function" && define.amd) {
    define(["require", "exports", "../release/go.js"], factory);
  } else {
    factory(null, window);
  }
})(function(require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    var go = require ? require("../release/go.js") : window.go;  // "go.js" had better already be loaded

    /**
    * @constructor
    * @param {string} name
    * @param {string | Diagram | null} diagram
    * @param {function(Test)} setup
    * @param {function(Test)} execute
    * @param {function(Test)} check
    * @class
    * A Test is a named object that can be {@link #run} to execute the test.
    * Tests are grouped in {@link TestCollection}s.
    * TestCollections and Tests form a tree structure, with Tests being the leaves.
    * <p/>
    * For convenience in defining tests, there are separate functions
    * for setting up a known state, for performing the actions that are being tested,
    * and for checking the side-effects of the test.
    * The setup functions are often shared by multiple Tests, and to a lesser extent,
    * so are the check functions.
    */
    function Test(name, diagram, setup, execute, check) {
      this.parent = null;
      this.name = name;
      this.diagramId = (typeof diagram === "string") ? diagram : "";
      this.diagram = (diagram instanceof go.Diagram) ? diagram : null;
      this.setup = setup;
      this.exec = execute;
      this.check = check;
      this.animated = false;
    }

    /**
    * Returns a string of names, starting with the root {@link TestCollection},
    * going down to this particular Test.
    * @this {Test}
    * @return {string}
    */
    Test.prototype.getPath = function() {
      var str = this.name;
      var t = this;
      while (t.parent) {
        str = t.parent.name + '|' + str;
        t = t.parent;
      }
      return str;
    };

   /**
    * reset current test to null, so the next test can run in TestCollection.run
    * @this {Test}
    */
    Test.prototype.unblock = function() {
      TestCollection.Current = null;
    }

    /**
    * Execute the three parts of each test: the setup, the execution, and the check.
    * @this {Test}
    */
    Test.prototype.run = function() {
      if (TestCollection.LogTestName) this.log(this.getPath());
      // find or reuse a Diagram
      if (this.diagramId !== "") {
        this.diagram = go.Diagram.fromDiv(this.diagramId);
        // don't reset() this!
      } else {
        if (this.diagram === null) {
          this.diagram = TestCollection.GlobalDiagram;
          // make sure it's hosted in the DOM
          if (this.diagram.div === null) {
            var div = document.getElementById("myDiagramDiv");
            if (!div) div = document.getElementById("myDiagram");
            this.diagram.div = div;
          }
        }
        this.diagram.reset();
        if ((new go.DiagramEvent).cancel === undefined) { // checks for 2.0
          if (this.diagram.reset === go.Diagram.prototype.reset) {
            var test = this;
            this.diagram.reset = function() {
              go.Diagram.prototype.reset.call(this);
              this.skipsUndoManager = true;
              this.initialContentAlignment = go.Spot.None; // forward compatible with 2.0
              this.animationManager.isEnabled = test.animated;
              this.skipsUndoManager = false;
            }
          }
        }
        if (!this.animated) {
          this.diagram.animationManager.stopAnimation();
          this.diagram.animationManager.isEnabled = false;
        } else {
          this.diagram.animationManager.isEnabled = true;
          // Tests MUST end with test.unblock(), unless we uncomment this
          // this.diagram.addDiagramListener('AnimationFinished', function() {
          //   TestCollection.Current = null; // reset this so the next test can run in TestCollection.run
          // })
        }
        var currentParent = this.parent;
        while (currentParent.parent) {
          currentParent = currentParent.parent;
          if (currentParent.preSetup) currentParent.preSetup(this);
        }
      }
      if (TestCollection.Catch) {
        try {
          if (this.setup) this.setup(this);
          if (this.diagram === null && this.diagramId !== "") {
            this.diagram = go.Diagram.fromDiv(this.diagramId);
          }
        } catch (ex) {
          this.error(ex.toString());
        }
        if (this.diagram !== null) { this.diagram.maybeUpdate(); }
        try {
          if (this.exec) this.exec(this);
        } catch (ex) {
          this.error(ex.toString());
        }
        if (this.diagram !== null) { this.diagram.maybeUpdate(); }
        try {
          if (this.check) this.check(this);
        } catch (ex) {
          this.error(ex.toString());
        }
      } else {
        if (this.setup) this.setup(this);
        if (this.diagram !== null) this.diagram.maybeUpdate();
        if (this.exec) this.exec(this);
        if (this.diagram !== null) this.diagram.maybeUpdate();
        if (this.check) this.check(this);
      }
      if (this.diagramId !== "") {
        if (this.diagram !== null) this.diagram.div = null;
      }
      if (this.savedEvent) {
        this.savedEvent = null; // events.js specific teardown
      }
    };

    /**
    * Log a message to the console, as a warning
    * @this {Test}
    * @param {string} msg
    */
    Test.prototype.log = function(msg) {
      if (window) {
        TestCollection.TotalWarnings++;
        if (window.console) window.console.log(msg);
      }
    };

    /**
    * Log an error message to the console, including this Test's full path/name.
    * If TestCollection.Alerts is true, also put up a JavaScript alert box.
    * @this {Test}
    * @param {string} msg
    */
    Test.prototype.error = function(msg) {
      if (window) {
        TestCollection.TotalErrors++;
        var str = this.getPath() + '\n  ' + msg;
        this.log(str);
        if (TestCollection.Alerts) window.alert(str);
      }
    };

    /**
    * This is the basic checking function.
    * @this {Test}
    * @param {boolean} pred
    * @param {string} msg
    */
    Test.prototype.assert = function(pred, msg) {
      if (!pred) {
        this.error(msg);
      }
    };

    Test.prototype.dumpAllNodeLocations = function() {
      var diagram = this.diagram;
      var datas = diagram.model.nodeDataArray;
      if (!window.console) return;
      var msg = this.name + ' [';
      for (var i = 0; i < datas.length; i++) {
        var node = diagram.findPartForData(datas[i]);
        var pos = node.location;
        if (i > 0) msg += ", ";
        msg += 'new go.Point(' + pos.x.toFixed(1) + ',' + pos.y.toFixed(1) + ')';
      }
      window.console.log(msg + ']');
    }

    /**
    * Dump the Points of all of the links.
    * @this {Test}
    */
    Test.prototype.dumpLinkPoints = function() {
      var diagram = this.diagram;
      if (window.console) {
        var datas = diagram.model.linkDataArray;
        for (var i = 0; i < datas.length; i++) {
          var link = diagram.findLinkForData(datas[i]);
          var points = link.points;
          var msg = '  [';
          for (var j = 0; j < points.length; j++) {
            if (j > 0) msg += ', ';
            msg += 'new go.Point(' + points.elt(j).x.toFixed(1) + ',' + points.elt(j).y.toFixed(1) + ')';
          }
          window.console.log(msg + ']');
        }
      }
    };

    /**
    * Dump the Geometries of all of the links.
    * @this {Test}
    */
    Test.prototype.dumpLinkGeometries = function() {
      var diagram = this.diagram;
      var datas = diagram.model.linkDataArray;
      for (var i = 0; i < datas.length; i++) {
        var link = diagram.findLinkForData(datas[i]);
        if (window.console) {
          var msg = this.name + ' ' + link.geometry.toString(1);
          window.console.log(msg);
        }
      }
    };

    // various useful methods for checking state

    /**
    * Check that all of the Nodes are at given locations.
    * @this {Test}
    * @param {Array} ptarr an Array of Points, the location of each Node corresponding to the data objects in model.nodeDataArray.
    */
    Test.prototype.assertAllNodeLocations = function(ptarr) {
      if (!Array.isArray(ptarr)) throw new Error("first arg is not an Array: " + ptarr);
      var model = this.diagram.model;
      var datas = model.nodeDataArray;
      for (var i = 0; i < ptarr.length; i++) {
        var expected = ptarr[i];
        var node = this.diagram.findNodeForData(datas[i]);
        var actual = node.location;
        this.assert(this.isApproxPoint(expected, actual), "Node is at location " + actual + " instead of at " + expected);
      }
    };

    /**
    * Check that all Links have specific routes.
    * This checks all Links in the order of the model.linkDataArray.
    * @this {Test}
    * @param {Array.Array.<Point>} testArrayPoints
    */
    Test.prototype.assertAllLinkPoints = function(testArrayPoints) {
      if (!Array.isArray(testArrayPoints)) throw new Error("first arg is not an Array: " + testArrayPoints);
      var model = this.diagram.model;
      var datas = model.linkDataArray;
      this.assert(Array.isArray(datas), "model.linkDataArray is not an Array: " + datas);
      for (var i = 0; i < datas.length; i++) {
        var data = datas[i];
        var link = this.diagram.findLinkForData(data);
        this.assert(link !== null, "Did not find Link for link data: " + data);
        this.assertLinkPoints(link, testArrayPoints[i]);
      }
    };

    /**
    * Check that a particular Link has a particular route, i.e. array of Points.
    * @this {Test}
    * @param {Link} link
    * @param {Array.<Point>} testPoints
    */
    Test.prototype.assertLinkPoints = function(link, testPoints) {
      if (!(link instanceof go.Link)) throw new Error("first arg is not a link: " + link);
      var linkPoints = link.points;
      if (!linkPoints || !testPoints) return;
      this.assert(linkPoints.length === testPoints.length,
        "The number of points of " + link + " is incorrect. Got: " +
        linkPoints.length + ". Expected: " + testPoints.length);
      if (linkPoints.length === testPoints.length) {
        var msg = "";
        for (var i = 0; i < testPoints.length; i++) {
          var linkPt = linkPoints.elt(i);
          var testPt = testPoints[i];
          if (!this.isApproxPoint(linkPt, testPt)) {
            msg += "\nPoint " + i + " of " + link + " has a value of (" + linkPt.x + ", " + linkPt.y +
              "), instead of (" + testPt.x + ", " + testPt.y + "), its expected value.";
          }
        }
        this.assert(msg.length === 0, msg);
      }
    };

    /**
    * Check the Geometry of every Link.
    * Unlike assertLinkPoints, this checks the path geometries, not just the Points of the routes.
    * @this {Test}
    * @param {Array.string} testArrayGeometries
    */
    Test.prototype.assertAllLinkGeometries = function(testArrayGeometries) {
      if (!Array.isArray(testArrayGeometries)) throw new Error("first arg is not an Array: " + testArrayGeometries);
      var model = this.diagram.model;
      var datas = model.linkDataArray;
      this.assert(Array.isArray(datas), "model.linkDataArray is not an Array: " + datas);
      for (var i = 0; i < datas.length; i++) {
        var data = datas[i];
        var link = this.diagram.findLinkForData(data);
        this.assert(link !== null, "Did not find Link for link data: " + data);
        this.assertLinkGeometry(link, testArrayGeometries[i]);
      }
    };

    /**
    * Check the Geometry of a single Link.
    * @this {Test}
    * @param {Link} link
    * @param {Geometry | string} testGeom
    */
    Test.prototype.assertLinkGeometry = function(link, testGeom) {
      var linkgeo = link.geometry;
      var geomstr = linkgeo.toString(1);
      var teststr = (typeof (testGeom) === 'string' ? testGeom : testGeom.toString(1));
      var testgeo = (testGeom instanceof go.Geometry ? testGeom : go.Geometry.parse(testGeom));
      var same = linkgeo.equalsApprox(testgeo) || geomstr === teststr;  //?? still depending on precise string comparisons when there are multiple PathFigures
      this.assert(same, "The geometry of " + link + " is incorrect. Got: " + geomstr + ".\n  Expected: " + teststr);
    }

    /**
    * Returns a list of every point that would be used in the drawing of the geometry
    * [x, y, x, y, x, y, x, y]
    * Making no distinction between control points or end points
    * Used for consistency/regression testing
    * @param {Geometry} geom
    * @this {Test}
    */
    Test.prototype.pathToPointArray = function(geom) {
      if (geom.type !== go.Geometry.Path) return null;
      var arr = [];
      var figs = geom.figures.iterator;
      // for each PathFigure:
      while (figs.next()) {
        var fig = figs.value;
        arr.push(fig.startX);
        arr.push(fig.startY);
        // for each segment of the PathFigure:
        var segments = fig.segments.iterator;
        var lastpointX = fig.startX;
        var lastpointY = fig.startY;
        while (segments.next()) {
          var seg = segments.value;
          switch (seg.type) {
            case go.PathSegment.Line:
            case go.PathSegment.Move:
              arr.push(seg.endX);
              arr.push(seg.endY);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case go.PathSegment.Bezier:
              arr.push(seg.point1X);
              arr.push(seg.point1Y);
              arr.push(seg.point2X);
              arr.push(seg.point2Y);
              arr.push(seg.endX);
              arr.push(seg.endY);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case go.PathSegment.QuadraticBezier:
              arr.push(seg.point1X);
              arr.push(seg.point1Y);
              arr.push(seg.endX);
              arr.push(seg.endY);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case go.PathSegment.Arc:
            case go.PathSegment.SvgArc:
              var curves = (seg.type === go.PathSegment.Arc) ? seg.buildBeziers(fig) : seg.buildSVGBeziers(fig, lastpointX, lastpointY);
              var clen = curves.length;
              var c = null;
              for (var k = 0; k < clen; k++) {
                c = curves[k];
                arr.push(c.x2);
                arr.push(c.y2);
                arr.push(c.x3);
                arr.push(c.y3);
                arr.push(c.x4);
                arr.push(c.y4);
              }
              if (c !== null) {
                lastpointX = c.x4;
                lastpointY = c.y4;
              }
              break;
            default:
              throw new Error('Unknown Segment type: ' + seg.type);
              break;
          }
        } // end segments
      } // end figures
      return arr;
    }

    /**
    * True if the difference between X and Y is less than 0.00000005.
    * @this {Test}
    * @param {number} x
    * @param {number} y
    * @return {boolean}
    */
    Test.prototype.isApproxEqual = function(x, y) {
      var d = x - y;
      return d < 0.00000005 && d > -0.00000005;
    };

    /**
    * True if the difference between X and Y is less than 0.5.
    * @this {Test}
    * @param {number} x
    * @param {number} y
    * @return {boolean}
    */
    Test.prototype.isApprox = function(x, y) {
      var d = x - y;
      return d < .5 && d > -.5;
    };

    /**
    * Check that both X and Y of each Point are approximately equal (< 0.5).
    * @this {Test}
    * @param {Point} a
    * @param {Point} b
    * @return {boolean}
    */
    Test.prototype.isApproxPoint = function(a, b) {
      return this.isApprox(a.x, b.x) && this.isApprox(a.y, b.y);
    };

    /**
    * Check that both Width and Height of each Size are approximately equal (< 0.5).
    * @this {Test}
    * @param {Size} a
    * @param {Size} b
    * @return {boolean}
    */
    Test.prototype.isApproxSize = function(a, b) {
      return this.isApprox(a.width, b.width) && this.isApprox(a.height, b.height);
    };

    /**
    * Check that X, Y, Width, and Height of each Rect are approximately equal (< 0.5).
    * @this {Test}
    * @param {Rect} a
    * @param {Rect} b
    * @return {boolean}
    */
    Test.prototype.isApproxRect = function(a, b) {
      return this.isApprox(a.x, b.x) && this.isApprox(a.y, b.y) &&
        this.isApprox(a.width, b.width) && this.isApprox(a.height, b.height);
    };

    // other useful functions

    /**
    * Simulate a mouse down event.
    * @this {Test}
    * @param {Point} pt a point in document coordinates.
    * @param {object=} eventprops an optional argument providing properties for the InputEvent.
    */
    Test.prototype.mouseDown = function(pt, eventprops, optionalDiagram) {
      var diagram = optionalDiagram || this.diagram;
      if (!diagram.isEnabled) return;
      if (eventprops && eventprops.sourceDiagram) diagram = eventprops.sourceDiagram;
      var n = new go.InputEvent();
      n.diagram = diagram;
      n.documentPoint = pt;
      n.viewPoint = diagram.transformDocToView(pt);
      n.down = true;
      if (eventprops) {
        for (var p in eventprops) {
          n[p] = eventprops[p];
          // If mac instead of control: true, switch to alt: true for drag-copy
          if (p === "control" && isMac) {
            n.alt = eventprops[p];
            n.meta = eventprops[p];
          }
        }
      }
      diagram.lastInput = n;
      diagram.firstInput = n.copy();
      diagram.currentTool.doMouseDown();
    };

    /**
    * Simulate a mouse move event.
    * @this {Test}
    * @param {Point} pt a point in document coordinates.
    * @param {object=} eventprops an optional argument providing properties for the InputEvent.
    */
    Test.prototype.mouseMove = function(pt, eventprops, optionalDiagram) {
      var diagram = optionalDiagram || this.diagram;
      if (!diagram.isEnabled) return;
      if (eventprops && eventprops.sourceDiagram) diagram = eventprops.sourceDiagram;
      var n = new go.InputEvent();
      n.diagram = diagram;
      n.documentPoint = pt;
      n.viewPoint = diagram.transformDocToView(pt);
      if (eventprops) {
        for (var p in eventprops) {
          n[p] = eventprops[p];
          // If mac instead of control: true, switch to alt: true for drag-copy
          if (p === "control" && isMac) {
            n.alt = eventprops[p];
            n.meta = eventprops[p];
          }
        }
      }
      diagram.lastInput = n;
      diagram.currentTool.doMouseMove();
    };

    /**
    * Simulate a mouse up event.
    * @this {Test}
    * @param {Point} pt a point in document coordinates.
    * @param {object=} eventprops an optional argument providing properties for the InputEvent.
    */
    Test.prototype.mouseUp = function(pt, eventprops, optionalDiagram) {
      var diagram = optionalDiagram || this.diagram;
      if (!diagram.isEnabled) return;
      if (eventprops && eventprops.sourceDiagram) diagram = eventprops.sourceDiagram;
      var n = new go.InputEvent();
      n.diagram = diagram;
      n.documentPoint = pt;
      n.viewPoint = diagram.transformDocToView(pt);
      n.up = true;
      if (diagram.firstInput.documentPoint.equals(pt)) n.clickCount = 1;  // at least??
      if (eventprops) {
        for (var p in eventprops) {
          n[p] = eventprops[p];
          // If mac instead of control: true, switch to alt: true for drag-copy
          if (p === "control" && isMac) {
            n.alt = eventprops[p];
            n.meta = eventprops[p];
          }
        }
      }
      diagram.lastInput = n;
      diagram.currentTool.doMouseUp();
    };

    /**
    * Simulate a mouse wheel event.
    * @this {Test}
    * @param {number} delta non-zero turn
    * @param {object=} eventprops an optional argument providing properties for the InputEvent.
    */
    Test.prototype.mouseWheel = function(delta, eventprops) {
      var diagram = this.diagram;
      if (!diagram.isEnabled) return;
      var n = diagram.lastInput.copy();
      n.diagram = diagram;
      n.delta = delta;
      if (eventprops) {
        for (var p in eventprops) {
          n[p] = eventprops[p];
          // If mac instead of control: true, switch to alt: true for drag-copy
          if (p === "control" && isMac) {
            n.alt = eventprops[p];
            n.meta = eventprops[p];
          }
        }
      }
      diagram.lastInput = n;
      diagram.currentTool.doMouseWheel();
    };

    /**
    * Simulate a key down event.
    * @this {Test}
    * @param {string|number} keyorcode
    * @param {object=} eventprops an optional argument providing properties for the InputEvent.
    */
    Test.prototype.keyDown = function(keyorcode, eventprops) {
      var diagram = this.diagram;
      if (!diagram.isEnabled) return;
      var n = diagram.lastInput.copy();
      n.diagram = diagram;
      if (typeof (keyorcode) === 'string') {
        n.key = keyorcode;
      } else if (typeof (keyorcode) === 'number') {
        n.key = String.fromCharCode(keyorcode);
      }
      n.down = true;
      if (eventprops) {
        for (var p in eventprops) {
          n[p] = eventprops[p];
          // If mac instead of control: true, switch to alt: true for drag-copy
          if (p === "control" && isMac) {
            n.alt = eventprops[p];
            n.meta = eventprops[p];
          }
        }
      }
      diagram.lastInput = n;
      diagram.currentTool.doKeyDown();
    };

    /**
    * Simulate a key up event.
    * @this {Test}
    * @param {string|number} keyorcode
    * @param {object=} eventprops an optional argument providing properties for the InputEvent.
    */
    Test.prototype.keyUp = function(keyorcode, eventprops) {
      var diagram = this.diagram;
      if (!diagram.isEnabled) return;
      var n = diagram.lastInput.copy();
      n.diagram = diagram;
      if (typeof (keyorcode) === 'string') {
        n.key = keyorcode;
      } else if (typeof (keyorcode) === 'number') {
        n.key = String.fromCharCode(keyorcode);
      }
      n.up = true;
      if (eventprops) {
        for (var p in eventprops) {
          n[p] = eventprops[p];
          // If mac instead of control: true, switch to alt: true for drag-copy
          if (p === "control" && isMac) {
            n.alt = eventprops[p];
            n.meta = eventprops[p];
          }
        }
      }
      diagram.lastInput = n;
      diagram.currentTool.doKeyUp();
    };


    /**
    * @constructor
    * @param {string} name
    * @class
    * A TestCollection is a named collection of {@link Test}s and "child" TestCollections.
    * TestCollections and Tests form a tree structure,
    * with TestCollection as the root and the internal nodes and with Tests being the leaf nodes.
    * <p/>
    * The TestRoot global variable holds the root TestCollection.
    */
    function TestCollection(name) {
      this.parent = null;
      this.name = name;
      this.tests = new go.List(/*Test*/);
      this.collections = new go.List(/*TestCollection*/);
    }

    /**
    * Add a {@link Test} or a {@link TestCollection} to this collection.
    * @this {TestCollection}
    * @param {Test|TestCollection} test
    */
    TestCollection.prototype.add = function(test) {
      if (test instanceof Test) {
        test.parent = this;
        this.tests.add(test);
      } else if (test instanceof TestCollection) {
        test.parent = this;
        this.collections.add(test);
      } else {
        throw new Error('not a Test or a TestCollection: ' + test);
      }
    };

    /**
    * Add a {@link Test} or a {@link TestCollection} to this collection.
    * @this {TestCollection}
    * @param {Test|TestCollection} test
    */
   TestCollection.prototype.addAnimated = function(test) {
    if (test instanceof Test) {
      test.parent = this;
      test.animated = true;
      this.tests.add(test);
    } else if (test instanceof TestCollection) {
      test.parent = this;
      this.collections.add(test);
    } else {
      throw new Error('not a Test or a TestCollection: ' + test);
    }
  };

    /**
    * Find a Test or a TestCollection that is directly in this collection.
    * @this {TestCollection}
    * @param {string} name
    * @return {Test|TestCollection}
    */
    TestCollection.prototype.find = function(name) {
      var it = this.tests.iterator;
      while (it.next()) {
        var t = it.value;
        if (t.name === name) return t;
      }
      it = this.collections.iterator;
      while (it.next()) {
        var tc = it.value;
        if (tc.name === name) return tc;
      }
      return null;
    };

    /**
    * Search for a Test with the given name, anywhere in the test hierarchy starting at this collection.
    * @this {TestCollection}
    * @param {string} name
    * @return {Test}
    */
    TestCollection.prototype.findTest = function(name) {
      var it = this.tests.iterator;
      while (it.next()) {
        var t = it.value;
        if (t.name === name) return t;
      }
      it = this.collections.iterator;
      while (it.next()) {
        var tc = it.value;
        var t = tc.findTest(name);
        if (t !== null) return t;
      }
      return null;
    }

    /**
    * Queue all of the tests in this collection, and all nested collections too.
    * @this {TestCollection}
    */
    TestCollection.prototype.queue = function() {
      var it = this.tests.iterator;
      while (it.next()) {
        var test = it.value;
        TestCollection.Queue.add(test);
      }
      var cit = this.collections.iterator;
      while (cit.next()) {
        var coll = cit.value;
        coll.queue();
      }
    };


    // Static methods and properties for controlling test runs

    /**
    * This static method starts running all Tests that are in the Queue.
    * It also puts out some statistics afterwards.
    */
    TestCollection.run = function() {
      if (window) {
        if (TestCollection.Queue.length > 0) {
          TestCollection.setStatus('Tests remaining: ' + TestCollection.Queue.length.toString());
          if (TestCollection.Current === null || (TestCollection.Current !== null && TestCollection.Current.animated === false)) {
            var first = TestCollection.Queue.elt(0);
            TestCollection.Queue.removeAt(0);
            TestCollection.Current = first;
            if (first) first.run();
          }
          var time = 1;
          if (TestCollection.Current !== null && TestCollection.Current.animated) time = 100;
          window.setTimeout(TestCollection.run, time);
        } else {
          var tim = '';
          if (TestCollection.StartTime) {
            var now = Date.now();
            var duration = now - TestCollection.StartTime;
            tim = ' in ' + (duration / 1000).toString() + ' seconds';
            TestCollection.StartTime = undefined;
          }
          var num = '';
          if (TestCollection.TotalNumber) num = TestCollection.TotalNumber.toString()
          var errs = '';
          if (TestCollection.TotalErrors) errs = ' with ' + TestCollection.TotalErrors.toString() + ' errors ';
          var warns = '';
          if (TestCollection.TotalWarnings) warns = ' and ' + TestCollection.TotalWarnings.toString() + ' warnings ';
          TestCollection.setStatus('finished running ' + num + ' tests ' + tim + errs);
          TestCollection.TotalNumber = 0;
          TestCollection.TotalErrors = 0;
        }
      }
    }

    /**
    * This static method updates a DIV named "myStatus" with a text message.
    * @param {string} msg
    */
    TestCollection.setStatus = function(msg) {
      var status = document.getElementById(TestCollection.StatusId);
      if (status !== null) {
        status.innerHTML = msg.toString();
      }
    };

    /**
    * This static method finds a Test given a path of names.
    * @param {string} path
    * @return {Test}
    */
    TestCollection.findTest = function(path) {
      var split = path.split('|');
      if (split.length === 1) {
        // search for a Test, anywhere in tree
        return TestRoot.findTest(split[0]);
      } else {
        var t = TestRoot;
        // ignore initial '|'
        for (var i = 1; i < split.length; i++) {
          var n = split[i];
          var c = t.find(n);
          if (c === null) return null;
          t = c;
        }
        return t;
      }
    };

    /**
    * This static method queues all tests in TestRoot, and then runs them all.
    */
    TestCollection.runAll = function() {
      TestCollection.StartTime = Date.now();
      TestCollection.TotalErrors = 0;
      TestRoot.queue();
      TestCollection.TotalNumber = TestCollection.Queue.length;
      TestCollection.run();
    };


    // Global variable:

    /**
    * The root of the hierarchy of defined Tests.
    * @type {TestCollection}
    */
    var TestRoot = new TestCollection('TestRoot');


    // Static variables for controlling how tests are run:

    TestCollection.GlobalDiagram = new go.Diagram();

    TestCollection.StatusId = 'myStatus';

    /**
    * The list of Tests to be run.
    * @type {List}
    */
    TestCollection.Queue = new go.List(/*Test*/);

    /**
    * Gets or sets whether very slow tests are run
    * @type {boolean}
    */
    TestCollection.RunSlowTests = false;

    /**
    * Gets or sets whether Test.run logs the name of the test.
    * @type {boolean}
    */
    TestCollection.LogTestName = true;

    /**
    * Gets or sets whether Test.error() will show an alert().
    * @type {boolean}
    */
    TestCollection.Alerts = true;

    /**
    * Gets or sets whether the test system will catch the exceptions from running tests.
    * Useful to turn off when debugging.
    * @type {boolean}
    */
    TestCollection.Catch = true;

    /**
    * Gets the currently running Test, if any.
    * @type {Test}
    */
    TestCollection.Current = null;

    // Static variables holding test results:

    /**
    * Keep track of how many Tests have been run by runAll.
    * @type {number}
    */
    TestCollection.TotalNumber = 0;

    /**
    * Keep track of how many errors occurred during runAll.
    * @type {number}
    */
    TestCollection.TotalErrors = 0;

    /**
    * Keep track of how many warnings occurred during runAll.
    * @type {number}
    */
    TestCollection.TotalWarnings = 0;


    // Initialize the tree of tests
    var myTreeDiagram = null;

    function getTree() {
      return myTreeDiagram;
    }

    function initTree(divid, statusid) {
      var $ = go.GraphObject.make;

      // this is the Diagram showing all of the tests
      myTreeDiagram = $(go.Diagram, divid ? divid : "myTestTree",
        { isModelReadOnly: true });

      // create Node template for each Test or TestCollection
      myTreeDiagram.nodeTemplateMap.add("TestCollection",
        $(go.Node, go.Panel.Horizontal,
          $(go.TextBlock,
            new go.Binding('text', '', go.Binding.toString),
            {
              click: function(e, obj) {
                var node = obj.part;
                var data = node.data;
                if (data instanceof Test) {
                  TestCollection.Queue.add(data);
                  TestCollection.TotalNumber = TestCollection.Queue.length;
                  TestCollection.run();
                } else if (data instanceof TestCollection) {
                  data.queue();
                  TestCollection.TotalNumber = TestCollection.Queue.length;
                  TestCollection.run();
                }
              }
            }),
          $("TreeExpanderButton")));

      myTreeDiagram.nodeTemplateMap.add("",  // Tests don't have any tree children
        $(go.Node, go.Panel.Horizontal,
          $(go.TextBlock,
            new go.Binding('text', '', go.Binding.toString),
            {
              click: function(e, obj) {
                var node = obj.part;
                var data = node.data;
                if (data instanceof Test) {
                  TestCollection.Queue.add(data);
                  TestCollection.TotalNumber = TestCollection.Queue.length;
                  TestCollection.run();
                } else if (data instanceof TestCollection) {
                  data.queue();
                  TestCollection.TotalNumber = TestCollection.Queue.length;
                  TestCollection.run();
                }
              }
            })));

      // define very simple Link template
      myTreeDiagram.linkTemplate =
        $(go.Link,
          { selectable: false },
          $(go.Shape,
            { isPanelMain: true }));

      // setup model data
      var nodedata = [];
      var linkdata = [];
      if (statusid) TestCollection.StatusId = statusid;
      walkTests(TestRoot, nodedata, linkdata);
      myTreeDiagram.model.nodeDataArray = nodedata;
      myTreeDiagram.model.linkDataArray = linkdata;

      // diagram layout
      myTreeDiagram.layout = $(go.TreeLayout, { nodeSpacing: 5 });
    }

    function walkTests(tc, nodes, links) {
      tc.category = "TestCollection";
      nodes.push(tc);
      tc.key = nodes.length;
      var it = tc.tests.iterator;
      while (it.next()) {
        var t = it.value;
        nodes.push(t);
        t.key = nodes.length;
        links.push({ from: tc.key, to: t.key });
      }
      it = tc.collections.iterator;
      while (it.next()) {
        var c = it.value;
        walkTests(c, nodes, links);
        links.push({ from: tc.key, to: c.key });
      }
    }

    function collapseTree(levels) {
      var before = Date.now();
      myTreeDiagram.startTransaction("collapsing tree");
      if (levels === undefined) levels = 3;
      var root = myTreeDiagram.findNodeForData(TestRoot);
      root.collapseTree(levels);
      var after = Date.now();
      var duration = after - before;
      myTreeDiagram.commitTransaction("collapsed tree");
      TestCollection.setStatus('collapsed tree in ' + (duration / 1000).toString() + ' seconds');
    }

    function expandTree() {
      myTreeDiagram.startTransaction("expanding tree");
      var root = myTreeDiagram.findNodeForData(TestRoot);
      root.expandTree(4);
      myTreeDiagram.commitTransaction("expanded tree");
    }

    exports.Test = Test;
    exports.TestCollection = TestCollection;
    exports.TestRoot = TestRoot;
    exports.initTree = initTree;
    exports.collapseTree = collapseTree;
    exports.getTree = getTree;
    exports.expandTree = expandTree;
});
